#include <string.h>
#include <stdio.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <linux/input.h>
#include <linux/uinput.h>
#include <stdio.h>
#include <sys/time.h>
#include <sys/types.h>
#include <unistd.h>
#include <sys/socket.h>
#include <arpa/inet.h> //inet_addr
#include <unistd.h>    //write

/* Global variables */
static int uinp_fd = -1;	 	// Joypad file descriptor
struct uinput_user_dev uinp; 	// Joypad device structure
struct input_event event; 		// Joypad device event structure
int connected = 0;

/* Setup the device */
int setup_uinput_device(){
	// Temporary variable
	int i=0;
	
	// Open the virtual input device
	uinp_fd = open("/dev/uinput", O_WRONLY | O_NDELAY);
	if (uinp_fd == 0){
		printf("Unable to open /dev/uinput\n");
		return -1;
	}

	// Assign name and properties to virtual device
	memset(&uinp,0,sizeof(uinp)); // Intialize the device to NULL
	strncpy(uinp.name, "Fishino Joystick", UINPUT_MAX_NAME_SIZE);
	uinp.id.version = 4;
	uinp.id.bustype = BUS_USB;

	// Setup the Joypad device
	ioctl(uinp_fd, UI_SET_EVBIT, EV_MSC);
	ioctl(uinp_fd, UI_SET_EVBIT, EV_KEY);
	ioctl(uinp_fd, UI_SET_EVBIT, EV_REL);
	ioctl(uinp_fd, UI_SET_EVBIT, EV_REP);
	for (i=0; i < 256; i++) {
		ioctl(uinp_fd, UI_SET_KEYBIT, i);
	}
	ioctl(uinp_fd, UI_SET_MSCBIT, MSC_SCAN);

	/* Create device into input sub-system */
	write(uinp_fd, &uinp, sizeof(uinp));
	if (ioctl(uinp_fd, UI_DEV_CREATE)){
		printf("Unable to create UINPUT device.");
		return -1;
	}

	return 1;
}



void send_key(char* key_str)
{
	int i = 0;
	uint16_t code;
	char keycode = key_str[0];
	
	// Joystick Configuration case 
	// (hold the 'ENTER' key for a bit time to enter configuration menu)
	if (keycode == '0'){
		code = KEY_ENTER;
		
		// Report BUTTON CLICK - PRESS event
		memset(&event, 0, sizeof(event));
		gettimeofday(&event.time, NULL);
		event.type = EV_MSC;
		event.code = MSC_SCAN;
		event.value = 32000;
		write(uinp_fd, &event, sizeof(event));
		event.type = EV_KEY;
		event.code = code;
		event.value = 1;
		write(uinp_fd, &event, sizeof(event));
		event.type = EV_SYN;
		event.code = SYN_REPORT;
		event.value = 0;
		write(uinp_fd, &event, sizeof(event));
	
		for (i=0; i< 10000; i++){
			memset(&event, 0, sizeof(event));
			gettimeofday(&event.time, NULL);
			event.type = EV_KEY;
			event.code = code;
			event.value = 2;
			write(uinp_fd, &event, sizeof(event));
			event.type = EV_SYN;
			event.code = SYN_REPORT;
			event.value = 0;
			write(uinp_fd, &event, sizeof(event));
		}
		
		// Report BUTTON CLICK - RELEASE event
		memset(&event, 0, sizeof(event));
		gettimeofday(&event.time, NULL);
		event.type = EV_MSC;
		event.code = MSC_SCAN;
		event.value = 32000;	//0x7004f
		write(uinp_fd, &event, sizeof(event));
		event.type = EV_KEY;
		event.code = code;
		event.value = 0;
		write(uinp_fd, &event, sizeof(event));
		event.type = EV_SYN;
		event.code = SYN_REPORT;
		event.value = 0;
		write(uinp_fd, &event, sizeof(event));
		
		return;
	}

	// Attention: only key mapped here will then be recognized by the Raspberry
	// (See Fishino firmware for correct key-mapping)
	switch(keycode){
		case 'a': code = KEY_A; break;
		case 's': code = KEY_S; break;
		case 'z': code = KEY_LEFT; break;
		case 'x': code = KEY_RIGHT; break;
		case 'c': code = KEY_UP; break;
		case 'v': code = KEY_DOWN; break;
		case 'k': code = KEY_ENTER; break;
		case 'l': code = KEY_SPACE; break;
		case 'q': code = KEY_Q; break;
		case 'w': code = KEY_W; break;
		case 'e': code = KEY_E; break;
		case 'r': code = KEY_R; break;
		case 't': code = KEY_T; break;
		case 'y': code = KEY_Y; break;
		case 'u': code = KEY_U; break;
		case 'i': code = KEY_I; break;
		case 'o': code = KEY_O; break;
		case 'p': code = KEY_P; break;
		case 'd': code = KEY_D; break;
		case 'f': code = KEY_F; break;
		case 'g': code = KEY_G; break;
		case 'h': code = KEY_H; break;
		case 'j': code = KEY_J; break;
		default: code= 0; break;
	}

	// Report BUTTON CLICK - PRESS event
	memset(&event, 0, sizeof(event));
	gettimeofday(&event.time, NULL);
	event.type = EV_MSC;
	event.code = MSC_SCAN;
	event.value = 32000;
	write(uinp_fd, &event, sizeof(event));
	event.type = EV_KEY;
	event.code = code;
	event.value = 1;
	write(uinp_fd, &event, sizeof(event));
	event.type = EV_SYN;
	event.code = SYN_REPORT;
	event.value = 0;
	write(uinp_fd, &event, sizeof(event));
	
	for (i=0; i< 1000; i++){
		// KEY HOLD event
		memset(&event, 0, sizeof(event));
		gettimeofday(&event.time, NULL);
		event.type = EV_KEY;
		event.code = code;
		event.value = 2;
		write(uinp_fd, &event, sizeof(event));
		event.type = EV_SYN;
		event.code = SYN_REPORT;
		event.value = 0;
		write(uinp_fd, &event, sizeof(event));
	}
	
	// Report BUTTON CLICK - RELEASE event
	memset(&event, 0, sizeof(event));
	gettimeofday(&event.time, NULL);
	event.type = EV_MSC;
	event.code = MSC_SCAN;
	event.value = 32000;
	write(uinp_fd, &event, sizeof(event));
	event.type = EV_KEY;
	event.code = code;
	event.value = 0;
	write(uinp_fd, &event, sizeof(event));
	event.type = EV_SYN;
	event.code = SYN_REPORT;
	event.value = 0;
	write(uinp_fd, &event, sizeof(event));
}


int main(int argc , char *argv[])
{
    int socket_desc , client_sock , c , read_size;
    struct sockaddr_in server , client;
    char client_message[10];
     
    //Create socket
    socket_desc = socket(AF_INET , SOCK_DGRAM , 0);
    if (socket_desc == -1)
    {
        printf("Could not create socket");
		return 1;
    }
    printf("Socket created");
     
    //Prepare the sockaddr_in structure
    server.sin_family = AF_INET;
    server.sin_addr.s_addr = INADDR_ANY;
    server.sin_port = htons( 8888 );
    
    int yes = 1;
    setsockopt(socket_desc, SOL_SOCKET, SO_REUSEADDR, &yes, sizeof(int));
    setsockopt(socket_desc, SOL_SOCKET, SO_REUSEPORT, &yes, sizeof(int));
 
    //Bind
    if( bind(socket_desc,(struct sockaddr *)&server , sizeof(server)) < 0)
    {
        //print the error message
        perror("bind failed. Error");
        return 1;
    }
    printf("bind done");
     
    //Listen
    listen(socket_desc , 1);
     
    //Accept and wait for incoming connection
    printf("Waiting for incoming connections...");
    c = sizeof(struct sockaddr_in);
     
    //Accept connection from an incoming client
    client_sock = accept(socket_desc, (struct sockaddr *)&client, (socklen_t*)&c);
    if (client_sock < 0)
    {
        perror("accept failed");
        return 1;
    }
    printf("Connection accepted");

	// Return an error if device not found.
	if (setup_uinput_device() < 0){
		printf("Unable to find uinput device\n");
		return -1;
	}

    connected = 1;
  
    //Receive a message from client
    while (connected && (read_size = recv(client_sock, client_message, 2, MSG_WAITALL) > 0)){
//		printf("Received key: %s\n", client_message);
	
		send_key(&client_message); // Send a key
	}
    
    if(read_size == 0)
    {
        puts("Client disconnected");
        fflush(stdout);
    }
    else if(read_size == -1)
    {
        perror("recv failed");
    }
	
	/* Destroy the device */
	ioctl(uinp_fd, UI_DEV_DESTROY);
	/* Close the Joypad device */
	close(uinp_fd);
     
    return 0;
}
